'''
Utility Functions for Challenges

This module defines useful and frequently used functions for use in Challenges.
Import this module to use them.
'''

from caboodle import web
from io import BytesIO
from PIL import Image
import base64
import requests
import pyscreenshot as ImageGrab

def _process_video_captcha(image):
	'''
	Processes Solve Media Video CAPTCHA

	Args:
		image (bytes): The overlay image encoded in base64

	Returns:
		A base64 encoded JPEG image

	The image processing includes cropping the overlay image to extract the
	bottom 20 pixels, adding a 10 pixel black border and scaling everything by a
	factor of 2.
	'''

	result = BytesIO()

	image = BytesIO(base64.b64decode(image))
	image = Image.open(image)
	image = image.convert('L').crop((0, 130, 300, 150)).crop((0, -10, 300, 30))
	image = image.resize((image.width * 2, image.height * 2), Image.ANTIALIAS)
	image.save(result, format = 'JPEG')

	return base64.b64encode(result.getvalue())

def _process_reveal_captcha(image):
	'''
	Processes Solve Media Reveal CAPTCHA

	Args:
		image (bytes): The overlay image encoded in base64

	Returns:
		A base64 encoded JPEG image

	The image processing includes adding a 10 pixel black border and scaling
	everything by a factor of 2.
	'''

	result = BytesIO()

	image = BytesIO(base64.b64decode(image))
	image = Image.open(image).crop((0, -10, 300, 30))
	image = image.resize((image.width * 2, image.height * 2), Image.ANTIALIAS)
	image.save(result, format = 'JPEG')

	return base64.b64encode(result.getvalue())

def get_element_image(element, browser):
	'''
	Takes a screenshot and crops out the element

	Args:
		element (WebElement): The element to crop
		browser (Browser): The web browser to use

	Returns:
		A base64 encoded JPEG image
	'''

	result = BytesIO()

	x, y = int(element.location['x']), int(element.location['y'])
	w, h = int(element.size['width']), int(element.size['height'])

	screen = BytesIO(base64.b64decode(browser.get_screenshot_as_base64()))
	image = Image.open(screen).crop((x, y, x + w, y + h))
	background = Image.new('RGB', image.size, (255, 255, 255))
	background.paste(image, mask = image.split()[3])
	background.save(result, format = 'JPEG')

	return base64.b64encode(result.getvalue())

def get_image_src(element):
	'''
	Downloads the source of an image

	Args:
		element (WebElement): The element to download

	Returns:
		A base64 encoded JPEG image
	'''

	result = BytesIO()

	image = BytesIO(requests.get(element.get_attribute('src')).content)
	Image.open(image).convert('L').save(result, format = 'JPEG')

	return base64.b64encode(result.getvalue())

def get_image_from_swf(url, timeout = 5):
	'''
	Takes a screenshot of a Flash object (.swf)

	Args:
		url (str): The URL of the SWF
		timeout (int): The length of time to wait before taking the screenshot

	Returns:
		A base64 encoded JPEG image
	'''

	result = BytesIO()

	browser = web.Browser()
	browser.set_timeout(timeout)
	browser.maximize_window()
	browser.get(url)
	bbox = browser.execute_script(
		'''
		var xy = [
			window.screenX + (window.outerWidth - window.innerWidth),
			window.screenY + (window.outerHeight - window.innerHeight)
		];
		var wh = [
			xy[0] + window.innerWidth,
			xy[1] + window.innerHeight
		];
		return xy.concat(wh);
		'''
	)
	ImageGrab.grab(bbox = bbox).save(result, format = 'JPEG')
	browser.quit()

	return base64.b64encode(result.getvalue())



'''
Solve Media CAPTCHA Challenges

This module is an implementation of the Challenge specification and collects
data to solve Solve Media CAPTCHAs. To use these Challenges, create a new
instance of them and call the `get_data()` function. Then, process the data
using an Agent and submit it by calling the `submit_data()` function. See the
unit tests for this module for more information.
'''

from caboodle.challenges.spec import Challenge
import caboodle.challenges.util as util
import time

class SolveMediaTextChallenge(Challenge):
	'''
	A Challenge for Solve Media Text CAPTCHAs

	A text CAPTCHA is the traditional type of CAPTCHA with obfuscated squiggly
	text. This Challenge will locate it if it exists and add a base64 encoded
	image of the CAPTCHA to the dictionary with the key 'image'. In addition to
	the CAPTCHA, the form to enter the result and the reload button to get a new
	CAPTCHA are added to the dictionary with the keys 'form' and 'reload'
	respectively.
	'''

	def __init__(self):
		super().__init__()

	def get_data(self, browser):
		'''
		Collects data needed to solve the Challenge

		Args:
			browser (Browser): The web browser to use

		Returns:
			A dictionary of collected data

		Raises:
			TypeError: The browser is not of type Browser
		'''

		super().get_data(browser)

		try:
			data = {}

			# Get elements
			data['image'] = browser.find_element_by_xpath(
				'//img[@id="adcopy-puzzle-image-image"]'
			)
			data['form'] = browser.find_element_by_id('adcopy_response')
			data['reload'] = browser.find_element_by_id('adcopy-link-refresh')

			# Preprocess elements
			data['image'] = util.get_element_image(data['image'], browser)

			return data
		except:
			return None

	def submit_data(self, data):
		'''
		Submits the processed data and solves the Challenge

		Args:
			data (dict): The Challenge to submit

		Raises:
			TypeError: The data is not a dictionary
		'''

		super().submit_data(data)

		data['form'].send_keys(data['result'])

class SolveMediaVideoChallenge(Challenge):
	'''
	A Challenge for Solve Media Video CAPTCHAs

	A video CAPTCHA is a unique type of CAPTCHA where the user must watch a
	short video before given the text to type in. This Challenge will locate the
	text if it exists and add a base64 encoded image of it to the dictionary
	with the key 'image'. In addition to the text, the form to enter the result
	and the reload button to get a new CAPTCHA are added to the dictionary with
	the keys 'form' and 'reload' respectively.
	'''

	def __init__(self):
		super().__init__()

	def get_data(self, browser):
		'''
		Collects data needed to solve the Challenge

		Args:
			browser (Browser): The web browser to use

		Returns:
			A dictionary of collected data

		Raises:
			TypeError: The browser is not of type Browser
		'''

		super().get_data(browser)

		try:
			data = {}

			# Play video
			browser.find_element_by_xpath('//img[@id="adcopy-poster"]').click()

			# Get elements
			while 'image' not in data:
				try:
					data['image'] = browser.find_element_by_xpath(
						'//img[@id="adcopy-ti-overlay"]'
					)
				except:
					time.sleep(1)

			data['form'] = browser.find_element_by_id('adcopy_response')
			data['reload'] = browser.find_element_by_id('adcopy-link-refresh')

			# Preprocess elements
			data['image'] = util.get_image_src(data['image'])
			data['image'] = util._process_video_captcha(data['image'])

			return data
		except:
			return None

	def submit_data(self, data):
		'''
		Submits the processed data and solves the Challenge

		Args:
			data (dict): The Challenge to submit

		Raises:
			TypeError: The data is not a dictionary
		'''

		super().submit_data(data)

		data['result'] = data['result'].split(' ')[-1:][0]
		data['form'].send_keys(data['result'])

class SolveMediaEmbedChallenge(Challenge):
	'''
	A Challenge for Solve Media Embed CAPTCHAs

	An embed CAPTCHA is a type of CAPTCHA where the user is presented with an
	embedded flash object and must enter the text it contains. This Challenge
	will locate it if it exists and add a base64 encoded image of it to the
	dictionary with the key 'image'. In addition to the image, the form to enter
	the result and the reload button to get a new CAPTCHA are added to the
	dictionary with the keys 'form' and 'reload' respectively.
	'''

	def __init__(self):
		super().__init__()

	def get_data(self, browser):
		'''
		Collects data needed to solve the Challenge

		Args:
			browser (Browser): The web browser to use

		Returns:
			A dictionary of collected data

		Raises:
			TypeError: The browser is not of type Browser
		'''

		super().get_data(browser)

		try:
			data = {}

			# Get SWF URL
			url = browser.find_element_by_xpath(
				'//embed[@id="adcopy-puzzle-image-image"]'
			).get_attribute('src')

			# Get image
			data['image'] = util.get_image_from_swf(url, 15)

			# Get elements
			data['form'] = browser.find_element_by_id('adcopy_response')
			data['reload'] = browser.find_element_by_id('adcopy-link-refresh')

			return data
		except:
			return None

	def submit_data(self, data):
		'''
		Submits the processed data and solves the Challenge

		Args:
			data (dict): The Challenge to submit

		Raises:
			TypeError: The data is not a dictionary
		'''

		super().submit_data(data)

		data['result'] = data['result'].replace(
			'Please enter the following:',
			''
		).replace(
			'Please enter:',
			''
		).strip()

		data['form'].send_keys(data['result'])

class SolveMediaRevealChallenge(Challenge):
	'''
	A Challenge for Solve Media Reveal CAPTCHAs

	A reveal CAPTCHA is a type of CAPTCHA where the user is blatantly given a
	text to type. This Challenge will locate it if it exists and add a base64
	encoded image of it to the dictionary with the key 'image'. In addition to
	the image, the form to enter the result and the reload button to get a new
	CAPTCHA are added to the dictionary with the keys 'form' and 'reload'
	respectively.
	'''

	def __init__(self):
		super().__init__()

	def get_data(self, browser):
		'''
		Collects data needed to solve the Challenge

		Args:
			browser (Browser): The web browser to use

		Returns:
			A dictionary of collected data

		Raises:
			TypeError: The browser is not of type Browser
		'''

		super().get_data(browser)

		try:
			data = {}

			# Enter iframe
			iframe = browser.find_element_by_xpath(
				'//iframe[starts-with(@id, "adcopy-unique-")]'
			)
			browser.switch_to_frame(iframe)

			# Get image
			data['image'] = browser.find_element_by_id('overlay')\
				.value_of_css_property('background-image')[27:-2]

			# Return to default context
			browser.switch_to_default_content()

			# Get elements
			data['form'] = browser.find_element_by_id('adcopy_response')
			data['reload'] = browser.find_element_by_id('adcopy-link-refresh')

			# Preprocess elements
			data['image'] = util._process_reveal_captcha(data['image'])

			return data
		except:
			return None

	def submit_data(self, data):
		'''
		Submits the processed data and solves the Challenge

		Args:
			data (dict): The Challenge to submit

		Raises:
			TypeError: The data is not a dictionary
		'''

		super().submit_data(data)

		data['result'] = data['result'].replace(
			'Please Enter:',
			''
		).replace(
			'Please Enter.',
			''
		).strip()

		data['form'].send_keys(data['result'])



from caboodle import web
from caboodle.challenges.recaptcha import RecaptchaV2Challenge
from caboodle.challenges.spec import Challenge
import unittest

class RecaptchaV2ChallengeTest(unittest.TestCase):
	'''
	Tests the RecaptchaV2Challenge to verify that it works correctly
	'''

	@classmethod
	def setUpClass(self):
		'''
		Creates a Challenge to test
		'''

		self.browser = web.Browser()
		self.challenge = RecaptchaV2Challenge()

	def test_init(self):
		'''
		Tests that the RecaptchaV2Challenge initializes correctly
		'''

		self.assertIsInstance(self.challenge, Challenge)

	def test_get_data(self):
		'''
		Tests that the RecaptchaV2Challenge can collect data
		'''

		self.assertFalse(self.challenge.get_data(self.browser))

		self.browser.get('https://www.google.com/recaptcha/api2/demo')

		data = self.challenge.get_data(self.browser)
		self.assertTrue('image' in data or data == 'solved')
		self.assertTrue('tiles' in data or data == 'solved')
		self.assertTrue('verify' in data or data == 'solved')
		self.assertTrue('text' in data or data == 'solved')
		self.assertTrue('tag' in data or data == 'solved')
		self.assertTrue('reload' in data or data == 'solved')
		self.assertTrue('rows' in data or data == 'solved')
		self.assertTrue('columns' in data or data == 'solved')

		with self.assertRaises(TypeError):
			self.challenge.get_data(None)

	@unittest.skip('Does not perform correctly, but function does work')
	def test_submit_data(self):
		'''
		Tests that the RecaptchaV2Challenge can submit data
		'''

		# Try until success or fail
		for _ in range(5):
			try:
				# Test valid data
				self.browser.get('https://www.google.com/recaptcha/api2/demo')

				data = self.challenge.get_data(self.browser)
				data['result'] = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9)

				self.challenge.submit_data(data)
				break
			except:
				pass
		else:
			self.fail('Could not submit valid data')

	@classmethod
	def tearDownClass(self):
		'''
		Closes the Browser to end the test
		'''

		self.browser.quit()

if __name__ == '__main__':
	unittest.main()



from caboodle import web
from caboodle.challenges.solve_media import SolveMediaTextChallenge
from caboodle.challenges.spec import Challenge
import os
import time
import unittest

HTML = None

# Locate html file
for root, dirs, files in os.walk(os.getcwd()):
	if 'solve_media.html' in files:
		HTML = 'file://' + os.path.join(root, 'solve_media.html')

class SolveMediaTextChallengeTest(unittest.TestCase):
	'''
	Tests the SolveMediaTextChallenge to verify that it works correctly
	'''

	@classmethod
	def setUpClass(self):
		'''
		Creates a Challenge to test
		'''

		self.browser = web.Browser()
		self.challenge = SolveMediaTextChallenge()

	def test_init(self):
		'''
		Tests that the SolveMediaTextChallenge initializes correctly
		'''

		self.assertIsInstance(self.challenge, Challenge)

	def test_get_data(self):
		'''
		Tests that the SolveMediaTextChallenge can collect data
		'''

		if HTML:
			self.assertFalse(self.challenge.get_data(self.browser))

			self.browser.get(HTML)

			# Wait for page to load
			while True:
				try:
					self.browser.find_element_by_id('show').click()
					break
				except:
					time.sleep(1)

			# Wait for CAPTCHA to load
			while True:
				try:
					self.browser.find_element_by_id('adcopy-puzzle-image-image')
					self.browser.find_element_by_id('adcopy_response')
					self.browser.find_element_by_id('adcopy-link-refresh')
					break
				except:
					time.sleep(1)

			data = self.challenge.get_data(self.browser)
			self.assertTrue(data)
			self.assertTrue(data['image'])
			self.assertTrue(data['form'])
			self.assertTrue(data['reload'])

			with self.assertRaises(TypeError):
				self.challenge.get_data(None)
		else:
			self.fail('Could not locate html file')

	def test_submit_data(self):
		'''
		Tests that the SolveMediaTextChallenge can submit data
		'''

		if HTML:
			self.browser.get(HTML)

			# Wait for page to load
			while True:
				try:
					self.browser.find_element_by_id('show').click()
					break
				except:
					time.sleep(1)

			# Wait for CAPTCHA to load
			while True:
				try:
					self.browser.find_element_by_id('adcopy-puzzle-image-image')
					self.browser.find_element_by_id('adcopy_response')
					self.browser.find_element_by_id('adcopy-link-refresh')
					break
				except:
					time.sleep(1)

			data = self.challenge.get_data(self.browser)
			print(data)
			data['result'] = '42'

			self.challenge.submit_data(data)

			self.assertEqual(data['form'].get_attribute('value'), '42')
		else:
			self.fail('Could not locate html file')

	@classmethod
	def tearDownClass(self):
		'''
		Closes the Browser to end the test
		'''

		self.browser.quit()

if __name__ == '__main__':
	unittest.main()
